#include "libs/extendedMemoryDriverMailbox.h"
#include <propeller.h>
#include "simpletools.h"

// Define global variables.
static extendedMemoryDriverMailbox_t *_pMemDriverMailbox;  // Global pointer to the memory driver mailbox.
static uint32_t _ioAccessorMask;
static uint32_t _flashCsMask;
static uint32_t _sRamCsMask = 1;
static uint32_t _clkMask;

void delay() {
  waitcnt(CNT + 8000000);
}

// Sends out the value stored in longToSend.
// MSB is shifted out first.
// numberOfBytesToSend defines, how many bytes of the long are sent.
void qpiSend(uint32_t longToSend, uint8_t numberOfBytesToSend) {
  //print("In qpiSend...\n");
  //print("longToSend: %b\n", longToSend);
  //print("Number of bytes to send from long: %d\n", numberOfBytesToSend);

  // Set all io pins to output
  DIRA |= _ioAccessorMask;
  //print("DIRA after setting up IO ports: %b\n", DIRA);

  // Copy longToSend.
  uint32_t longToSendClone = longToSend;

  uint32_t i;
  for (i = 0; i < (2 * numberOfBytesToSend); i++) {
    // Set clk pin low
    OUTA &= ~_clkMask;
    //print("Setting clk port to low: %b\n", OUTA);

    // Clear io pin values
    OUTA &= ~_ioAccessorMask;

    // Set the values of the 4 most left bits of longToSend to the io pins.
    // To be able to do that, we have to bring the 4 most left
    // bits of longToSendClone into the position where the IO pins on the OUTA
    // port are.
    longToSendClone >>= 28;
    
    // Now shift the values into place.
    longToSendClone <<= _pMemDriverMailbox->io0Pin;
    //print("longToSendClone after bringing the most left 4 bits into io port place: %b\n", longToSendClone);

    OUTA |= longToSendClone;
    //print("OUTA after or with longToSendClone: %b\n", OUTA);

    delay();
    // Set clk pin high
    OUTA |= _clkMask;
    //print("Setting clk port to high: %b\n", OUTA);

    longToSend <<= 4;
    longToSendClone = longToSend;
    delay();
  }
}

void qpiReceive(uint32_t readByteCount, uint8_t *pDestinationAddress) {
  // Set all io pins to input
  //print("In qpiReceive...\n");
  DIRA &= ~_ioAccessorMask;
  //print("Setting IO Ports to input, DIRA: %b\n", DIRA);

  uint32_t i;
  for (i = 0; i < readByteCount; i++) {
    // Set clock to low. This forces the flash to send the data to the io pins.
    OUTA &= ~_clkMask;
    //print("Setting clock to low\n");
    delay();
    // Set clock pin to high.
    OUTA |= _clkMask;
    //print("Setting clock to high");
  
    delay();
    // Read in the data.
    uint32_t inData = INA;
    
    // TODO: Remove the following line after debugging finished!
    //inData = 11;
    //print("inData after reading from port: %b\n", inData);

    // Now shift the inData to the right until the 4
    // read bits are at the most right position of the long.
    inData >>= _pMemDriverMailbox->io0Pin;

    // Now shift the data 4 bits to the left.
    // This makes the read 4 bits the ms bits of the byte.
    inData <<= 4;
    //print("inData after bringing read 4 bits into msb position: %b\n", inData);

    // Set clock pin to low to force flash to write out data.
    delay();
    OUTA &= ~_clkMask;
    //print("Setting clock to low\n");

    delay();
    // Set clock pin to high again
    OUTA |= _clkMask;
    //print("Setting clock to high");

    delay();
    // Read in data
    uint32_t inData2 = INA;

    // TODO: Remove the following line after debugging finished!
    //inData2 = 15;
    //print("inData2 after reading from port: %b\n", inData2);

    inData2 >>= _pMemDriverMailbox->io0Pin;

    // Now clear every bit but the last four bits
    inData2 &= 0xF;
    //print("inData2 after clearing the most left 4 bits: %b\n", inData2);

    // Combine inData and inData2.
    inData |= inData2;
    //print("inData after combining: %b\n", inData);
    //printf("inData in dec: %d\n", inData);

    // Write the read byte to the destination address.
    *(pDestinationAddress + i) = (uint8_t) inData;
    delay();
  }
}

void spiSend(uint32_t longToSend, uint8_t numberOfBytesToSend) {
  uint32_t i;
  uint8_t j;
  uint32_t bitMask = 1 << 31;
  uint32_t outMask = 1 << _pMemDriverMailbox->io0Pin;
  //print("longToSend: %b\n", longToSend);
  
  uint32_t ioMask = 1 << _pMemDriverMailbox->io0Pin;
  // Set io0Pin (PMC DI Pin) to output
  DIRA |= ioMask;

  // Set io1 to input
  ioMask = 1 << (_pMemDriverMailbox->io0Pin + 1);
  DIRA &= ~ioMask;
  //print("Direction after setting IO0 to output, IO1 to input: %b\n", DIRA);

  for (i = 0; i < numberOfBytesToSend; i++) {
    for (j = 0; j < 8; j++) {
      // Set clock to low
      OUTA &= ~_clkMask;

      // Set pin state to the value of the msb of longToSend
      uint32_t msbValue = longToSend & bitMask;
      if (msbValue < 1) {
        // Clear bit
        OUTA &= ~outMask;
      } else {
        // Set it if bit is 1
        OUTA |= outMask;
      }
      //print("OUTA after setting IO0 pin to value: %b\n", OUTA);

      // Set clock pin to high
      OUTA |= _clkMask;

      // Shift longToSend one bit to the left
      longToSend <<= 1;
      delay();
    }
  }
}

void spiReceive(uint32_t readByteCount, uint8_t *pDestinationAddress) {
  print("In spiReceive...\n");
  uint32_t i;
  uint8_t j;
  uint8_t readByte = 0;
  uint32_t spiDoMask = 1 << (_pMemDriverMailbox->io0Pin + 1);
  print("spiDoMask: %b\n", spiDoMask);
  
  for (i = 0; i < readByteCount; i++) {
    for (j = 0; j < 8; j++) {
      delay();
      // Set clk pin state to low which forces chip to deliver data
      OUTA &= ~_clkMask;
      //print("Clock to low, OUTA: %b\n", OUTA);
      
      delay();
      // Set clock pin to high
      OUTA |= _clkMask;
      //print("Clock to high, OUTA: %b\n", OUTA);
      //print("INA state: %b\n", INA);
      //print("INA & spiDoMask = %d\n", INA & spiDoMask);

      // Read spi DO pin state
      if ((INA & spiDoMask) > 0) {
        // Bit set
        readByte |= 1;
      } else {
        // Bit not set
        readByte &= ~1;
      }

      //print("readByte after setting most right bit: %b\n", readByte);
      // Shift readByte to bit to the left
      if (j < 7) {
        readByte <<= 1;
      }
    }

    //print("readByte after casting to byte: %x\n", readByte);
    *(pDestinationAddress + i) = readByte;
    readByte = 0;
  }
}

void selectSRAM() {
  OUTA &= ~_sRamCsMask;
  delay();
}

void releaseSRAM() {
  OUTA |= _sRamCsMask;
  delay();
}

void initSRAM() {
  // Set sequential mode
  uint32_t mode = 0x01400000;
  selectSRAM();
  spiSend(mode, 2);
  releaseSRAM();

  // Set to QPI mode
  mode = 0x38000000;
  selectSRAM();
  spiSend(mode, 1);
  releaseSRAM();
}

void testSRAM() {
  // Write a long to SRAM
  uint32_t command = 0x02000000;  // Write sram beginning at address 0
  uint32_t longToSend = 0xaabbccdd;
  uint8_t readBuffer[4];

  selectSRAM();
  // Write command at address 0
  qpiSend(command, 4);
  print("Start sending long value...\n");
  qpiSend(longToSend, 4);
  releaseSRAM();

  // Now read back value
  command = 0x03000000;   // Read sram beginning at address 0
  selectSRAM();
  qpiSend(command, 4);
  
  // Send dummy byte
  qpiSend(0, 1);

  // Start reading
  qpiReceive(4, &(readBuffer[0]));
  releaseSRAM();

  print("readBuffer0 = %x\n", readBuffer[0]);
  print("readBuffer1 = %x\n", readBuffer[1]);
  print("readBuffer2 = %x\n", readBuffer[2]);
  print("readBuffer3 = %x\n", readBuffer[3]);
}

void readFlashJedecId() {
  uint8_t jedecResult[3];
  jedecResult[0] = 0;
  jedecResult[1] = 0;
  jedecResult[2] = 0;

  // Bring flash into QPI mode
  uint32_t command = 0x38000000;
  // Set flash chip select to low
  OUTA &= ~_flashCsMask;
  spiSend(command, 1);
  // Set flash chip select to high (disable chip)
  OUTA |= _flashCsMask;

  // Read JEDEC ID via QPI read
  command = 0x9F000000;
  // Set flash chip select to low
  OUTA &= ~_flashCsMask;
  qpiSend(command, 1);
  qpiReceive(3, &(jedecResult[0]));
  // Set flash chip select to high (disable chip)
  OUTA |= _flashCsMask;

  print("jedecResult[0]: %x\n", jedecResult[0]);
  print("jedecResult[1]: %x\n", jedecResult[1]);
  print("jedecResult[2]: %x\n", jedecResult[2]);

  // Exit QPI mode
  command = 0xFF000000;
  OUTA &= ~_flashCsMask;
  qpiSend(command, 1);
  // Set flash chip select to high (disable chip)
  OUTA |= _flashCsMask;
}

void initAll() {
  // Setup the io pin accessor mask
  _ioAccessorMask = 15 << _pMemDriverMailbox->io0Pin;
  print("ioAccessorMask: %b\n", _ioAccessorMask);

  // Setup flash cs mask
  _flashCsMask = 1 << _pMemDriverMailbox->flashCSPin;
  print("flashCsMask: %b\n", _flashCsMask);

  // Setup sram cs mask
  _sRamCsMask = 1 << _pMemDriverMailbox->sramCSPin;
  print("sRamCsMask: %b\n", _sRamCsMask);

  // Setup clk mask
  _clkMask = 1 << _pMemDriverMailbox->clkPin;
  print("clkMask: %b\n", _clkMask);

  // Set all io pins to input
  DIRA &= ~_ioAccessorMask;

  // Set clk and chip select pins to output;
  DIRA |= _flashCsMask;
  DIRA |= _sRamCsMask;
  DIRA |= _clkMask;

  // Set all chip select values to high (disabled) and the clk pin to low
  OUTA |= _flashCsMask;
  OUTA |= _sRamCsMask;
  OUTA &= ~_clkMask;
}

void main (extendedMemoryDriverMailbox_t *pMemDriverMailbox) {
  pMemDriverMailbox->io0Pin = 0;
  pMemDriverMailbox->flashCSPin = 5;
  pMemDriverMailbox->sramCSPin = 6;
  pMemDriverMailbox->clkPin = 7;

  _pMemDriverMailbox = pMemDriverMailbox;

  initAll();
  initSRAM();
  testSRAM();
  //readFlashJedecId();
  print("Finished...\n");
}